import decimal
import sys
from collections import namedtuple
from typing import TypedDict, Callable, TypeVar, Sequence
from datetime import date, datetime
import asyncio

from graphql import GraphQLError
from graphql.type.definition import GraphQLObjectType, GraphQLEnumType, get_named_type, is_list_type, is_non_null_type
from ariadne.types import GraphQLResolveInfo
from sqlalchemy import Row, ColumnElement, select, Executable, cast, String, Select, and_
from sqlalchemy.orm import DeclarativeBase, aliased
from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy.exc import MultipleResultsFound
from pyparsing import ParseException

from .auth_utils import User
from .models_utils import Relation
from .graphql_app import check_permission, HideResult
from .filter_parser import number_filter_parser, date_filter_parser, datetime_filter_parser, boolean_filter_parser

M = TypeVar('M', bound=DeclarativeBase)
nulldate = date(1753, 1, 1)  # For Incadea Tables


class FieldFilter(TypedDict):
    field_name: str
    value: str


# Shortcuts

def get_table_primary_keys(model: type[M]) -> list[ColumnElement]:
    return [col for col in model.__mapper__.columns if col.primary_key]


def filter_primary_key(q: Executable, model: type[M], primary_key_values: list[str]):
    pk_cols = get_table_primary_keys(model)
    try:
        for i, pk_val in enumerate(primary_key_values):
            q = q.where(pk_cols[i] == pk_val)
    except IndexError:
        raise GraphQLError(f"You have specified more primary keys values than there are columns for model {model.__name__}")
    return q


def add_typename_to_rows(model: type[DeclarativeBase], rows: Sequence[Row], type_name=None) -> list[namedtuple]:
    if type_name is None:
        type_name = model.__name__
    keys = [col.key or col.name for col in model.__mapper__.columns]
    NamedTuple = namedtuple(type_name, keys)
    NamedTuple.__typename = type_name
    return [NamedTuple(*(getattr(row, key) for key in keys)) for row in rows]


def get_user(info: GraphQLResolveInfo) -> User:
    return info.context['request'].user


def get_db(info) -> AsyncConnection:
    return info.context['request'].state.db


# Todo: remove column and just use model with should be a selectable
async def get_or_none(db: AsyncConnection, model: type[M], primary_keys: list, column: ColumnElement | None = None) -> M | None:
    result = await db.execute(filter_primary_key(select(column or model), model, primary_keys))
    if column is not None:
        return result.scalar_one_or_none()
    else:
        return result.one_or_none()


async def get_or_error(db: AsyncConnection, model: type[M], primary_keys: list, column: ColumnElement | None = None) -> M:
    try:
        row: M | None = await get_or_none(db, model, primary_keys, column)
    except MultipleResultsFound:
        raise GraphQLError(f"Primary keys: {primary_keys} returned more than one row")
    if row is None:
        raise GraphQLError(f"Could not find {model.name}: {primary_keys}")
    return row


# Resolver Tools

def separate_filters(filters: list[FieldFilter], field_names_to_separate: list[str]):
    """ When some filters are automatically handled, and others you need to write custom SQLAlchemy queries """
    newfilters = []
    separated = []
    for f in filters:
        if f['field_name'] in field_names_to_separate:
            separated.append(f)
        else:
            newfilters.append(f)
    return newfilters, separated


def model_to_dict(model, remove_fields: list[str] = None):  # Todo: Maybe remove since switching to core
    if remove_fields is None:
        remove_fields = []
    return {
        field: getattr(model, field)
        for field in model.__mapper__.column_attrs.keys()
        if field not in remove_fields
    }


# Complete Resolvers

def resolve_type_inspector_factory(model_base: type[DeclarativeBase]):
    models_dict = {model.class_.__name__: model.class_ for model in model_base.registry.mappers}
    # Todo: Deduce filtering automatically
    def resolve_type_inspector(_, info: GraphQLResolveInfo, type_name: str):
        gqltype = info.schema.get_type(type_name)
        if gqltype is None or not isinstance(gqltype, GraphQLObjectType):
            return None

        # Primary Keys. Raise error when using directive and model not found, else ignore
        model_name = getattr(gqltype, '__modelname__', type_name)
        model: type[DeclarativeBase] | None = models_dict.get(model_name)
        if model is not None:
            primary_keys = [col.name for col in get_table_primary_keys(model)]
        elif hasattr(gqltype, '__modelname__'):
            raise GraphQLError(f"Could not find model with name {model_name}")
        else:
            primary_keys = None

        # Filters
        all_filter = hasattr(gqltype, '__all_filter__')
        field_details = []
        for field_name, field in gqltype.fields.items():
            has_filter = False
            if hasattr(field, '__filter__'):
                if getattr(field, '__filter__'):
                    has_filter = True
            elif all_filter:
                has_filter = True

            field_filter_type = None
            if has_filter:
                field_type = get_named_type(field.type)
                if field_type is None:
                    raise Exception('Can only filter on Named Types')
                # Deducing filter type by GraphQL type. Contrary to simple_table_resolver
                if is_list_type(field.type) or (is_non_null_type(field.type) and is_list_type(field.type.of_type)):
                    field_filter_type = 'LIST'  # If list, it means it's a postgresql array and only = comparator works
                elif field_type.name == 'String':
                    field_filter_type = 'STRING'
                elif field_type.name in ['Int', 'Float']:
                    field_filter_type = 'NUMBER'
                elif field_type.name in ['Date', 'DateTime']:
                    field_filter_type = 'DATE'
                elif field_type.name == 'Boolean':
                    field_filter_type = 'BOOLEAN'
                elif isinstance(field_type, GraphQLEnumType):
                    field_filter_type = 'STRING'  # Consider Enum as strings
                else:
                    raise GraphQLError(f'Type {field_type.name} cannot support filtering on field {field_name}')

            # Todo: implement editable
            field_details.append({'field_name': field_name, 'filter_type': field_filter_type, 'editable': False})

        return {'field_details': field_details, 'primary_keys': primary_keys}
    return resolve_type_inspector


def apply_model_relation(q: Select, model: type[DeclarativeBase], relation_name: str) -> tuple[Select, type[DeclarativeBase]]:
    try:
        relation: Relation = getattr(model, '__relations__', {})[relation_name]
    except KeyError:
        raise GraphQLError(f"Could not find Relation {relation_name} in model {model.__name__}")

    rel_model = aliased(relation.model)

    if relation.join_on is None:
        q = q.join(rel_model)
    else:
        q = q.join(rel_model, and_(*(
            getattr(rel_model, foreign_col) == local_col
            for foreign_col, local_col in relation.join_on.items())
        ))

    return q, rel_model


def load_from_model_query(model: type[M], filters: list[FieldFilter], limit: int | None, offset: int | None,
                          query_modifier: Callable[[Executable], Executable] | None = None, init_query: Executable | None = None) -> Executable:
    q = select(model) if init_query is None else init_query

    for f in filters:
        full_name = f['field_name']
        value = f['value']

        *relation_names, field_name = full_name.split('.')
        current_model = model
        for relation_name in relation_names:
            # Get Relation model and join
            q, current_model = apply_model_relation(q, current_model, relation_name)

        field = getattr(current_model, field_name)
        field_type = field.type.python_type

        # Deducing filter type by model column type. Contrary to resolve_type_inspector.
        try:
            if field_type is str:
                q = q.where(cast(field, String).ilike(value))  # cast used to make Enum behave like strings.
            elif field_type in [int, float, decimal.Decimal]:
                q = number_filter_parser(q, field, value)
            elif field_type is date:
                q = date_filter_parser(q, field, value)
            elif field_type is datetime:
                q = datetime_filter_parser(q, field, value)
            elif field_type is bool:
                q = boolean_filter_parser(q, field, value)
            elif field_type is list:
                q = q.where(field.any_() == value)
            else:
                raise GraphQLError(f'Cannot filter on column type {field_type}')
        except ParseException as e:
            raise GraphQLError(f'Cannot parse value: {value} for field {field} of type {field_type} [{e}]')

    if query_modifier is not None:
        q = query_modifier(q)

    return q.limit(limit).offset(offset)


def resolve_type_loader_factory(model_base: type[DeclarativeBase]):
    models_dict = {model.class_.__name__: model.class_ for model in model_base.registry.mappers}

    async def resolve_type_loader(_, info, type_name, filters: list[FieldFilter], limit: int, offset: int):
        gqltype = info.schema.get_type(type_name)
        if gqltype is None:  # Check if Type exists in GQL
            raise GraphQLError(f'Type {type_name} does not exist')
        model_name = getattr(gqltype, '__modelname__', type_name)

        try:
            model = models_dict[model_name]
        except KeyError:
            raise GraphQLError(f"Could not find {model_name} in Models")

        q = load_from_model_query(model, filters, limit, offset)
        rows = (await get_db(info).execute(q)).all()
        return add_typename_to_rows(model, rows, type_name)

    return resolve_type_loader


def simple_table_resolver_factory(model: type[M], query_modifiers=None):
    async def simple_table_resolver(_, info, filters: list[FieldFilter], limit: int, offset: int) -> list[M]:
        q = load_from_model_query(model, filters, limit, offset, query_modifiers)
        return (await get_db(info).execute(q)).all()
    return simple_table_resolver


async def external_module_executor(module_name, *args: str):
    proc = await asyncio.create_subprocess_exec(sys.executable, '-u', '-m', f'scripts.{module_name}', *args,
                                                stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
    while not proc.stdout.at_eof():
        data = await proc.stdout.readline()
        yield data.decode().rstrip()

    error = await proc.stderr.read()
    if error:
        raise GraphQLError(error.decode().rstrip())


def subscription_permission_check(generator):
    async def new_generator(obj, info, *args, **kwargs):
        try:
            check_permission(info)
        except HideResult:
            yield None
            return

        async for res in generator(obj, info, *args, **kwargs):
            yield res

    return new_generator



from typing import Any, Mapping
import asyncio
from datetime import date, datetime
from importlib import resources
from graphql import GraphQLField, GraphQLObjectType, GraphQLInterfaceType, GraphQLResolveInfo
from ariadne import load_schema_from_path, make_executable_schema, SchemaDirectiveVisitor, FallbackResolversSetter, \
    ScalarType, UnionType
from ariadne.graphql import GraphQLError
from ariadne.asgi import GraphQL
from ariadne.asgi.handlers import GraphQLHTTPHandler, GraphQLTransportWSHandler
from ariadne.types import Extension, SchemaBindable

from .auth_utils import get_token_from_header, get_user_from_token, JWTAuthBackend, AuthCredentials


# Directives

class NoAuthenticationDirective(SchemaDirectiveVisitor):
    def visit_field_definition(
            self,
            field: GraphQLField,
            object_type: GraphQLObjectType | GraphQLInterfaceType,
    ) -> GraphQLField:
        field.__require_authentication__ = False
        return field

    def visit_object(self, object_: GraphQLObjectType) -> GraphQLObjectType:
        object_.__require_authentication__ = False
        return object_


class NeedPermissionDirective(SchemaDirectiveVisitor):
    def visit_field_definition(
        self,
        field: GraphQLField,
        object_type: GraphQLObjectType | GraphQLInterfaceType,
    ) -> GraphQLField:
        if self.args['strict'] or object_type.name == 'Mutation' or object_type.name == 'Subscription':
            field.__require_scope__ = self.args['scope']
        else:
            field.__hide_noscope__ = self.args['scope']
        return field

    def visit_object(self, object_: GraphQLObjectType) -> GraphQLObjectType:
        if self.args['strict']:
            object_.__require_scope__ = self.args['scope']
        else:
            object_.__hide_noscope__ = self.args['scope']
        return object_


class DBModelDirective(SchemaDirectiveVisitor):
    def visit_object(self, object_: GraphQLObjectType) -> GraphQLObjectType:
        object_.__modelname__ = self.args['model_name']
        return object_


class AllFilterDirective(SchemaDirectiveVisitor):
    def visit_object(self, object_: GraphQLObjectType) -> GraphQLObjectType:
        object_.__all_filter__ = True
        return object_


class FilterDirective(SchemaDirectiveVisitor):
    def visit_field_definition(
        self,
        field: GraphQLField,
        object_type: GraphQLObjectType | GraphQLInterfaceType,
    ) -> GraphQLField:
        field.__filter__ = True
        return field


class NoFilterDirective(SchemaDirectiveVisitor):
    def visit_field_definition(
        self,
        field: GraphQLField,
        object_type: GraphQLObjectType | GraphQLInterfaceType,
    ) -> GraphQLField:
        field.__filter__ = False
        return field


# Middleware

class HideResult(BaseException):
    pass


def check_permission(info: GraphQLResolveInfo):
    if info.field_name != '__schema':
        request = info.context['request']
        field = info.parent_type.fields[info.field_name]

        # Check for Authentication
        if hasattr(field, '__require_authentication__'):
            needs_auth = field.__require_authentication__
        elif hasattr(info.parent_type, '__require_authentication__'):
            needs_auth = info.parent_type.__require_authentication__
        else:
            needs_auth = True

        if needs_auth and not request.user.is_authenticated:
            raise GraphQLError(message='Requires Authentication')

        # check for Strict Permission
        if hasattr(field, '__require_scope__'):
            needs_scope = field.__require_scope__
        elif hasattr(info.parent_type, '__require_scope__'):
            needs_scope = info.parent_type.__require_scope__
        else:
            needs_scope = None

        if needs_scope is not None and needs_scope not in request.auth.scopes:
            raise GraphQLError(message=f'Requires Scope: {needs_scope}')

        # check for Loose Permission
        if hasattr(field, '__hide_noscope__'):
            hide_noscope = field.__hide_noscope__
        elif hasattr(info.parent_type, '__hide_noscope__'):
            hide_noscope = info.parent_type.__hide_noscope__
        else:
            hide_noscope = None

        if hide_noscope is not None and hide_noscope not in request.auth.scopes:
            raise HideResult()


async def check_permission_middleware(resolver, obj, info: GraphQLResolveInfo, **args):
    try:
        check_permission(info)
    except HideResult:
        return None

    # Return resolver
    if asyncio.iscoroutinefunction(resolver):
        return await resolver(obj, info, **args)
    else:
        return resolver(obj, info, **args)


class DBRollbackExtension(Extension):
    def has_errors(self, errors, context) -> None:
        context['request'].state.errors = errors


# Custom Scalars


model_types = UnionType('ModelTypes')


@model_types.type_resolver
def resolve_model_types(obj, *_):
    return obj.__typename


date_scalar = ScalarType("Date")


@date_scalar.serializer
def serialize_date(value: date | datetime):
    if isinstance(value, datetime):
        value = value.date()
    return value.isoformat()


@date_scalar.value_parser
def parse_date(value: str):
    return date.fromisoformat(value)


datetime_scalar = ScalarType("DateTime")


@datetime_scalar.serializer
def serialize_datetime(value: datetime):
    if type(value) is date:
        raise GraphQLError("Cannot convert Date Type into DateTime")
    return value.isoformat()


@datetime_scalar.value_parser
def parse_datetime(value: str):
    return date.fromisoformat(value)


# Fallback Resolver

def strict_default_field_resolver(source: Any, info: GraphQLResolveInfo, **args: Any) -> Any:
    """Strict Default field resolver.
    Same as default but raise Error when atrribute doesn't exist instead of returning None
    """
    field_name = info.field_name
    try:
        value = (
            source[field_name]
            if isinstance(source, Mapping)
            else getattr(source, field_name)
        )
    except (KeyError, AttributeError):
        raise GraphQLError(f"Attribute {field_name} does not exist on {info.parent_type.name} object")

    if callable(value):
        return value(info, **args)
    return value


class StrictFallbackResolverSetter(FallbackResolversSetter):
    def add_resolver_to_field(self, _: str, field_object: GraphQLField) -> None:
        if field_object.resolve is None:
            field_object.resolve = strict_default_field_resolver


def on_websocket_connect(websocket, payload: Any):
    """ Alternative to Header authentication if cookie auth cannot be used """
    if not isinstance(payload, Mapping) or 'Authorization' not in payload:
        return
    token = get_token_from_header(payload['Authorization'])

    for middleware in websocket.scope['app'].user_middleware:
        auth_backend = middleware.options.get('backend')
        if isinstance(auth_backend, JWTAuthBackend):
            user = get_user_from_token(token, auth_backend.secret)
            websocket.scope['user'] = user
            websocket.scope['auth'] = AuthCredentials(user.scopes)
            break


# Exported

def create_graphql(schema_path, bindables: list[SchemaBindable], debug: bool):
    fallback_resolvers = StrictFallbackResolverSetter()
    type_defs = load_schema_from_path(schema_path)
    with resources.path('st_microservice', 'base.graphql') as common_filename:
        type_defs_common = load_schema_from_path(common_filename)

    schema = make_executable_schema([type_defs, type_defs_common], bindables, model_types, date_scalar, datetime_scalar, fallback_resolvers, directives={
        'no_authentication': NoAuthenticationDirective,
        'need_permission': NeedPermissionDirective,
        'db_model': DBModelDirective,
        'all_filter': AllFilterDirective,
        'filter': FilterDirective,
        'no_filter': NoFilterDirective
    })

    return GraphQL(
        schema,
        debug=debug,
        http_handler=GraphQLHTTPHandler(
            middleware=[check_permission_middleware],
            extensions=[DBRollbackExtension]
        ),
        websocket_handler=GraphQLTransportWSHandler(on_connect=on_websocket_connect)
    )



from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.exc import SQLAlchemyError
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint


class AsyncDBMiddleware(BaseHTTPMiddleware):
    """ Will start a DB Session at every request and commit or rollback in the end """
    def __init__(self, app, database_uri: str):
        super().__init__(app)
        self.engine = create_async_engine(database_uri, connect_args={'timeout': 5})

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        async with self.engine.connect() as connection:
            request.state.db = connection

            # Continue with request
            response = await call_next(request)

            if hasattr(request.state, 'errors'):
                await connection.rollback()
            else:
                try:  # Try to commit
                    await connection.commit()
                except SQLAlchemyError:
                    await connection.rollback()
                    return JSONResponse({'errors': 'Error while commiting to Database'}, status_code=500)  # Todo: adopt graphql spec
            return response



from os import getenv
from starlette.applications import Starlette
from starlette.routing import BaseRoute, Route
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.responses import JSONResponse

from .database import AsyncDBMiddleware
from .auth_utils import JWTAuthBackend, auth_error_handler


# Utilities
def get_required_env(var_name: str) -> str:
    var = getenv(var_name)
    if var is None:
        raise EnvironmentError(f'Could not get {var_name} from ENV')
    return var


def hasenv(var_name: str) -> bool:
    return getenv(var_name, False) is not False


def get_user(request) -> JSONResponse:
    """ Get user route """
    return JSONResponse(request.user.to_json() if request.user.is_authenticated else None)


# App factory
def create_app(routes: list[BaseRoute], secret: str, root_domain: str, database_uri: str | None, debug=False) -> Starlette:
    authbackend = JWTAuthBackend(secret)

    # Allow origins from both HTTP or HTTPS, root or subdomains, any port
    root_domain_re = r'https?://(.*\.)?{}(:\d*)?'.format(root_domain.replace('.', r'\.'))
    print('Allowed Origins Regex:', root_domain_re)

    middleware = [
        Middleware(CORSMiddleware, allow_origin_regex=root_domain_re, allow_credentials=True, allow_methods=['*']),
        Middleware(AsyncDBMiddleware, database_uri=database_uri),
        Middleware(AuthenticationMiddleware, backend=authbackend, on_error=auth_error_handler)
    ]

    app = Starlette(debug=debug, routes=routes+[Route('/getuser', get_user, methods=['GET'])], middleware=middleware)
    app.router.redirect_slashes = False
    return app



from typing import Any
import jwt
from starlette.authentication import BaseUser, AuthenticationBackend, AuthenticationError, AuthCredentials
from starlette.requests import HTTPConnection
from starlette.responses import JSONResponse


class User(BaseUser):
    def __init__(self, user_id: str, display_name: str, scopes: list[str]):
        self.user_id = user_id
        self._display_name = display_name
        self.scopes = scopes

    @property
    def display_name(self) -> str:
        return self._display_name

    @property
    def is_authenticated(self):
        return True

    @property
    def identity(self) -> str:
        return self.user_id

    def to_json(self) -> dict[str, Any]:
        return {'user_id': self.user_id, 'display_name': self.display_name, 'scopes': self.scopes}


def get_token_from_header(authorization: str) -> str:
    try:
        scheme, token = authorization.split()
    except ValueError:
        raise AuthenticationError('Could not separate Authorization scheme and token')
    if scheme.lower() != 'bearer':
        raise AuthenticationError(f'Authorization scheme {scheme} is not supported')
    return token


def get_user_from_token(token: str, secret: str) -> User:
    try:
        payload = jwt.decode(token, secret, algorithms=['HS256'])
        user = User(payload['user_id'], payload['display_name'], payload['scopes'])
        return user
    except jwt.InvalidTokenError as e:
        raise AuthenticationError(f'JWT Token error: {e}')
    except Exception as e:
        raise AuthenticationError(f'Authentication error: {e}')


class JWTAuthBackend(AuthenticationBackend):
    def __init__(self, secret: str):
        self.secret = secret

    async def authenticate(self, request: HTTPConnection) -> tuple[AuthCredentials, User] | None:
        # Try to get token from header then from cookies
        auth_header = request.headers.get('Authorization', None)
        if auth_header is not None:
            token = get_token_from_header(auth_header)
        else:
            token = request.cookies.get('token')

        if token is None:
            return None

        user = get_user_from_token(token, self.secret)

        return AuthCredentials(user.scopes), user


def auth_error_handler(request, exc: Exception):
    response = JSONResponse({'error': str(exc)}, 400)
    response.delete_cookie('token')
    return response
